const fs = require('fs')
const path = require('path')
const LRU = require('lru-cache')
const express = require('express')
const compression = require('compression')
const {createBundleRenderer} = require('vue-server-renderer')
const SSRBuilder = require('flow-build')

const port = process.env.PORT || 8080

const app = express()

const resolve = file => path.resolve(__dirname, file)

const isProd = process.env.NODE_ENV === 'production'

const templatePath = resolve('./src/index.template.html')


const serve = (path, cache) => express.static(resolve(path), {
  maxAge: cache && isProd ? 1000 * 60 * 60 * 24 * 30 : 0
})

let renderer

function createRenderer(mfs = fs) {
  const template = fs.readFileSync(templatePath, 'utf-8')
  let distPath = path.resolve(process.cwd(), 'dist')

  const bundle = JSON.parse(mfs.readFileSync(path.resolve(distPath, './server-bundle.json'), "utf-8"))
  const clientManifest = JSON.parse(mfs.readFileSync(path.resolve(distPath, './vue-ssr-client-manifest.json'), "utf-8"));

  let option = {
    cache: false,
    basedir: resolve('./dist'),
    runInNewContext: false,
    template,
    clientManifest
  }

  if(isProd) {
    option.cache = LRU({
      max: 1000,
      maxAge: 1000 * 60 * 15  // 15 mins
    })
  }

  renderer = createBundleRenderer(bundle, option)
  return renderer
}

const render = (req, res) => {
  const s = Date.now()

  res.setHeader("Content-Type", "text/html")

  const handleError = err => {
    if (err.url) {
      res.redirect(err.url)
    } else if (err.code === 404) {
      res.status(404).send('404 | Page Not Found')
    } else {
      // Render Error Page or Redirect
      res.status(500).send('500 | Internal Server Error')
      console.error(`error during render : ${req.url}`)
      console.error(err.stack)
    }
  }

  const context = {
    title: 'Vue SSR', // default title
    url: req.url
  }

  renderer.renderToString(context, (err, html) => {
    if (err) {
      return handleError(err)
    }
    res.send(html)
    if (!isProd) {
      console.log(`whole request: ${Date.now() - s}ms`)
    }
  })
}

const launch = (cb) => {
  // request
  app.get('*', render)

  app.listen(port, () => {
    console.log(`server started at localhost:${port}`)
    if(cb) {
      cb()
    }
  })
}


app.use(compression({threshold: 0}))
app.use('/static', serve('./dist/static', false))


if (isProd) {
  renderer = createRenderer()
  launch()
} else {
  const builder = new SSRBuilder(require('./flow.config'))
  builder.build(createRenderer).then(({devMiddleware, hotMiddleware}) => {
    app.use(devMiddleware)
    app.use(hotMiddleware)
    launch(() => {
      builder.openBrowser("localhost", port)
    })
  }).catch((err) => {
    console.log(err)
  })
}
